# Copyright 2004-2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

__metaclass__ = type

import urllib
import cgi

from zope.interface import implements, classProvides
from zope.cachedescriptors.property import Lazy

from lazr.batchnavigator.z3batching.batch import _Batch
from lazr.batchnavigator.interfaces import (
    IBatchNavigator, InvalidBatchSizeError, IBatchNavigatorFactory)

__all__ = ['BatchNavigator']

class BatchNavigator:

    # subclasses can override
    _batch_factory = _Batch

    implements(IBatchNavigator)
    classProvides(IBatchNavigatorFactory)

    start_variable_name = 'start'
    batch_variable_name = 'batch'

    default_batch_size = 50
    max_batch_size = 300

    # We want subclasses to be able to hide the 'Last' link from
    # users.  They may want to do this for really large result sets;
    # for example, batches with over a hundred thousand items.
    show_last_link = True

    # The default heading describing the kind of objects in the batch.
    # Sub-classes can override this to be more specific.
    default_singular_heading = 'result'
    default_plural_heading = 'results'

    transient_parameters = None

    @Lazy
    def query_string_parameters(self):
        query_string = self.request.get('QUERY_STRING', '')

        # Just in case QUERY_STRING is in the environment explicitly as
        # None (Some tests seem to do this, but not sure if it can ever
        # happen outside of tests.)
        if query_string is None:
            query_string = ''
        return cgi.parse_qs(query_string, keep_blank_values=True)

    def __init__(self, results, request, start=0, size=None, callback=None,
                 transient_parameters=None, force_start=False):
        "See `IBatchNavigatorFactory.__call__`"
        self.request = request
        local = (self.batch_variable_name, self.start_variable_name)
        self.transient_parameters = set(local)
        if transient_parameters is not None:
            self.transient_parameters.update(transient_parameters)

        # For backwards compatibility (as in the past a work-around has been
        # to include the url batch params in hidden fields within posted
        # forms), if the request is a POST request, and either the 'start'
        # or 'batch' params are included then revert to the default behaviour
        # of using the request (which automatically gets the params from the
        # request.form dict).
        if request.method == 'POST' and (
            self.start_variable_name in request.form or
            self.batch_variable_name in request.form):
            batch_params_source = request
        else:
            # We grab the request variables directly from the requests
            # query_string_parameters so that they will be recognized
            # even during post operations.
            batch_params_source = dict(
                (k, v[0]) for k, v
                in self.query_string_parameters.items() if k in local)

        # In this code we ignore invalid request variables since it
        # probably means the user finger-fumbled it in the request. We
        # could raise UnexpectedFormData, but is there a good reason?
        request_start = batch_params_source.get(
            self.start_variable_name, None)
        if force_start or request_start is None:
            self.start = start
        else:
            try:
                self.start = int(request_start)
            except (ValueError, TypeError):
                self.start = start

        self.default_size = size

        request_size = batch_params_source.get(self.batch_variable_name, None)
        if request_size:
            try:
                size = int(request_size)
            except (ValueError, TypeError):
                pass
            if size > self.max_batch_size:
                raise InvalidBatchSizeError(
                    'Maximum for "%s" parameter is %d.' %
                    (self.batch_variable_name,
                     self.max_batch_size))

        if size is None:
            size = self.default_batch_size

        self.batch = self._batch_factory(results, start=self.start, size=size)
        if callback is not None:
            callback(self, self.batch)
        self.setHeadings(
            self.default_singular_heading, self.default_plural_heading)

    @property
    def heading(self):
        """See `IBatchNavigator`"""
        if self.batch.total() == 1:
            return self._singular_heading
        return self._plural_heading

    def setHeadings(self, singular, plural):
        """See `IBatchNavigator`"""
        self._singular_heading = singular
        self._plural_heading = plural

    def getCleanQueryString(self, params=None):
        """Removes start and batch params if present and returns a query
        string.

        If ``params`` is None, uses the current query_string_params.
        """
        if params is None:
            params = []
            for k, v in self.query_string_parameters.items():
                params.extend((k, item) for item in v)
        else:
            try:
                params = params.items()
            except AttributeError:
                pass

        # We need the doseq=True because some url params are for multi-value
        # fields.
        return urllib.urlencode(
            [(key, value) for (key, value) in sorted(params)
             if key not in self.transient_parameters],
             doseq=True)

    def generateBatchURL(self, batch):
        url = ""
        if not batch:
            return url

        qs = self.getCleanQueryString()
        if qs:
            qs += "&"

        start = batch.startNumber() - 1
        size = batch.size
        base_url = str(self.request.URL)
        url = "%s?%s%s=%d" % (base_url, qs, self.start_variable_name, start)
        if size != self.default_size:
            # The current batch size should only be part of the URL if it's
            # different from the default batch size.
            url = "%s&%s=%d" % (url, self.batch_variable_name, size)
        return url

    def getBatches(self):
        batch = self.batch.firstBatch()
        batches = [batch]
        while 1:
            batch = batch.nextBatch()
            if not batch:
                break
            batches.append(batch)
        return batches

    def firstBatchURL(self):
        batch = self.batch.firstBatch()
        if self.start == 0:
            # We are already on the first batch.
            batch = None
        return self.generateBatchURL(batch)

    def prevBatchURL(self):
        return self.generateBatchURL(self.batch.prevBatch())

    def nextBatchURL(self):
        return self.generateBatchURL(self.batch.nextBatch())

    def lastBatchURL(self):
        batch = self.batch.lastBatch()
        if self.start == batch.start:
            # We are already on the last batch.
            batch = None
        return self.generateBatchURL(batch)

    def batchPageURLs(self):
        batches = self.getBatches()
        urls = []
        size = len(batches)

        nextb = self.batch.nextBatch()

        # Find the current page
        if nextb:
            current = nextb.start/nextb.size
        else:
            current = size

        self.current = current
        # Find the start page to show
        if (current - 5) > 0:
            start = current-5
        else:
            start = 0

        # Find the last page to show
        if (start + 10) < size:
            stop = start + 10
        else:
            stop = size

        initial = start
        while start < stop:
            this_batch = batches[start]
            url = self.generateBatchURL(this_batch)
            if (start+1) == current:
                urls.append({'['+str(start + 1)+']' : url})
            else:
                urls.append({start + 1 : url})
            start += 1

        if current != 1:
            url = self.generateBatchURL(batches[0])
            urls.insert(0, {'_first_' : url})
        if current != size:
            url = self.generateBatchURL(batches[size-1])
            urls.append({'_last_':url})

        return urls

    def currentBatch(self):
        return self.batch



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
# Some parts Copyright 2005 Canonical Ltd.
#
##############################################################################
"""
Zope-derived Batching Support

*Do not use this. It is now an internal implementation detail.*
The intent is to get rid of this in favor of moving to extending z3c.batching.
"""


from zope.interface import implements
from zope.interface.common.sequence import IFiniteSequence
from interfaces import IBatch

from zope.cachedescriptors.property import Lazy

# The base batch size, which can be overridden by users of _Batch such
# as BatchNavigator. In Launchpad, we override it via a config option.
BATCH_SIZE = 50

class _Batch(object):
    implements(IBatch)

    def _get_length(self, results):
        # LBYL: we don't want to mask exceptions that len might raise, and we
        # don't have chained exceptions yet.
        if getattr(results, '__len__', None) is None:
            results = IFiniteSequence(results)
        return len(results)

    def __init__(self, results, start=0, size=None, _listlength=None):
        if results is None:
            results = []
        self.list = results

        if _listlength is None:
            _listlength = self ._get_length(results)
        self.listlength = _listlength

        if size is None:
            size = BATCH_SIZE
        self.size = size

        if start >= _listlength:
            self.trueSize = 0
        elif start+size >= _listlength:
            self.trueSize = _listlength-start
        else:
            self.trueSize = size

        if _listlength == 0:
            start = -1

        self.start = start
        if self.trueSize == 0:
            self.end = start
        else:
            self.end = start+self.trueSize-1

    def __len__(self):
        if self.trueSize < 0:
            return 0
        else:
            return self.trueSize

    def __getitem__(self, key):
        if key >= self.trueSize:
            raise IndexError, 'batch index out of range'
        # When self.start is negative (IOW, when we are batching over an
        # empty list) we need to raise IndexError here; otherwise, the
        # attempt to slice below will, on objects which don't implement
        # __len__, raise a mysterious exception directly from python.
        if self.start < 0:
            raise IndexError, 'batch is empty'
        return self.list[self.start+key]

    @Lazy
    def sliced_list(self):
        # We use Lazy here to avoid self.__iter__ giving us new objects every
        # time; in certain cases (such as when database access is necessary)
        # this can be expensive.
        return list(self.list[self.start:self.end+1])

    def __iter__(self):
        if self.start < 0:
            # as in __getitem__, we need to avoid slicing with negative
            # indices in this code
            return iter([])
        return iter(self.sliced_list)

    def first(self):
        return self[0]

    def last(self):
        return self[self.trueSize-1]

    def __contains__(self, item):
        return item in self.__iter__()

    def nextBatch(self):
        start = self.start + self.size
        if start >= self.listlength:
            return None
        return _Batch(self.list, start, self.size, _listlength=self.listlength)

    def prevBatch(self):
        # The only case in which we should /not/ offer a previous batch
        # is when we are already at position zero, which also happens
        # when the list is empty.
        if self.start == 0:
            return None
        if self.start > self.listlength:
            return self.lastBatch()
        else:
            start = self.start - self.size
        if start < 0:
            # This situation happens, for instance, when you have a
            # 20-item batch and you manually set your start to 15;
            # in this case, hopping back one batch would be starting at
            # -5, which doesn't really make sense.
            start = 0
        return _Batch(self.list, start, self.size, _listlength=self.listlength)

    def firstBatch(self):
        return _Batch(self.list, 0, size=self.size,
                      _listlength=self.listlength)

    def lastBatch(self):
        # Return the last possible batch for this dataset, at the
        # correct offset.
        last_index = self.listlength - 1
        last_batch_start = last_index - (last_index % self.size)
        if last_batch_start < 0:
            last_batch_start = 0
        return _Batch(self.list, last_batch_start, size=self.size,
                      _listlength=self.listlength)

    def total(self):
        return self.listlength

    def startNumber(self):
        return self.start+1

    def endNumber(self):
        return self.end+1




# Copyright 2004-2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

from zope.interface import Interface, Attribute

__metaclass__ = type


class InvalidBatchSizeError(AssertionError):
    """Received a batch parameter that exceed our configured max size."""

class IBatchNavigator(Interface):
    """A batch navigator for a specified set of results."""

    batch = Attribute("The IBatch for which navigation links are provided.")

    heading = Attribute(
        "The heading describing the kind of objects in the batch.")

    def setHeadings(singular, plural):
        """Set the heading for singular and plural results."""

    def prevBatchURL():
        """Return a URL to the previous chunk of results."""

    def nextBatchURL():
        """Return a URL to the next chunk of results."""

    def batchPageURLs():
        """Return a list of links representing URLs to pages of
        results."""

class IBatchNavigatorFactory(Interface):
    def __call__(results, request, start=0, size=None, callback=None,
                 transient_parameters=None, force_start=False):
        """Constructs an `IBatchNavigator` instance.

        :param results: is an iterable of results.

        :param request: will be inspected for a start variable; if set,
            it indicates which point we are currently displaying at. It
            will also be inspected for a batch variable; if set, it will
            be used instead of the size supplied in the callsite.  Expected to
            confirm to zope.publisher.interfaces.IRequest.

        :param size: is the default batch size, to fall back to if the
            request does not specify one.  If no size can be determined
            from arguments or request, the launchpad.default_batch_size
            config option is used.

        :param callback: is called, if defined, at the end of object
            construction with the defined batch as determined by the
            start and request parameters.

        :param transient_parameters: optional sequence of parameter
            names that should not be passed on in links generated by
            the batch navigator.  Use this for parameters that had
            meaning when this page was requested, but should not be kept
            around for other page requests in the batch.

        :param force_start: if True, the given `start` argument has
            precedence over the start value in the request, if any.

        :raises InvalidBatchSizeError: if the requested batch size is higher
            than the maximum allowed.
        """



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Batch tests.
"""
import unittest

from lazr.batchnavigator.z3batching.batch import _Batch
from lazr.batchnavigator.z3batching.interfaces import IBatch

class BatchTest(unittest.TestCase):

    def getData(self):
        return ['one', 'two', 'three', 'four', 'five', 'six',
                'seven', 'eight', 'nine', 'ten']

    def test_Interface(self):
        self.failUnless(IBatch.providedBy(_Batch([], 0, 0)))

    def test_constructor(self):
        _Batch(self.getData(), 9, 3)
        # A start that's larger than the length should still construct a working
        # (but empty) batch.
        _Batch(self.getData(), start=99, size=3)

    def test__len__(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(len(batch), 3)
        batch = _Batch(self.getData(), 9, 3)
        self.assertEqual(len(batch), 1)
        batch = _Batch(self.getData(), 99, 3)
        self.assertEqual(len(batch), 0)

    def test__getitem__(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(batch[0], 'one')
        self.assertEqual(batch[1], 'two')
        self.assertEqual(batch[2], 'three')
        batch = _Batch(self.getData(), 3, 3)
        self.assertEqual(batch[0], 'four')
        self.assertEqual(batch[1], 'five')
        self.assertEqual(batch[2], 'six')
        batch = _Batch(self.getData(), 9, 3)
        self.assertRaises(IndexError, batch.__getitem__, 3)

    def test__iter__(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(list(iter(batch)), ['one', 'two', 'three'])
        batch = _Batch(self.getData(), 9, 3)
        self.assertEqual(list(iter(batch)), ['ten'])
        batch = _Batch(self.getData(), 99, 3)
        self.assertEqual(list(iter(batch)), [])

    def test__contains__(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assert_(batch.__contains__('one'))
        self.assert_(batch.__contains__('two'))
        self.assert_(batch.__contains__('three'))
        self.assert_(not batch.__contains__('four'))
        batch = _Batch(self.getData(), 6, 3)
        self.assert_(not batch.__contains__('one'))
        self.assert_(batch.__contains__('seven'))
        self.assert_(not batch.__contains__('ten'))

    def test_firstBatch(self):
        """Check that the link to the first batch works.

        This first batch will be always pointing to the first available batch
        and, its main difference with the 'prev' and 'next' batches is, that
        will not be None ever.
        """
        # first batch when we are at the beginning of the batch.
        first = _Batch(self.getData(), 0, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we are in the second set of items of the batch.
        first = _Batch(self.getData(), 3, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we are in the third set of items of the batch.
        first = _Batch(self.getData(), 6, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we are at the end of the batch.
        first = _Batch(self.getData(), 9, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we get a request for an out of range item.
        first = _Batch(self.getData(), 99, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])

    def test_nextBatch(self):
        next = _Batch(self.getData(), 0, 3).nextBatch()
        self.assertEqual(list(iter(next)), ['four', 'five', 'six'])
        nextnext = next.nextBatch()
        self.assertEqual(list(iter(nextnext)), ['seven', 'eight', 'nine'])
        next = _Batch(self.getData(), 9, 3).nextBatch()
        self.assertEqual(next, None)
        next = _Batch(self.getData(), 99, 3).nextBatch()
        self.assertEqual(next, None)

    def test_prevBatch(self):
        prev = _Batch(self.getData(), 9, 3).prevBatch()
        self.assertEqual(list(iter(prev)), ['seven', 'eight', 'nine'])
        prevprev = prev.prevBatch()
        self.assertEqual(list(iter(prevprev)), ['four', 'five', 'six'])
        prev = _Batch(self.getData(), 0, 3).prevBatch()
        self.assertEqual(prev, None)
        prev = _Batch(self.getData(), 2, 3).prevBatch()
        self.assertEqual(list(iter(prev)), ['one', 'two', 'three'])
        last = _Batch(self.getData(), 99, 3).prevBatch()
        self.assertEqual(list(iter(last)), ['ten'])

    def test_lastBatch(self):
        """Check that the link to the last batch works.

        This last batch will be always pointing to the last available batch
        and, its main difference with the 'prev' and 'next' batches is, that
        will not be None ever.
        """
        # last batch when we are at the beginning of the batch.
        last = _Batch(self.getData(), 0, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we are in the second set of items of the batch.
        last = _Batch(self.getData(), 3, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we are in the third set of items of the batch.
        last = _Batch(self.getData(), 6, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we are at the end of the batch.
        last = _Batch(self.getData(), 9, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we get a request for an out of range item.
        last = _Batch(self.getData(), 99, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])

        # We are going to test now the same, but when we get a request of 5
        # items per batch because we had a bug in the way we calculate the
        # last batch set that was only happening when we were using a batch
        # size that is multiple of the item list length.

        # last batch when we are at the beginning of the batch.
        last = _Batch(self.getData(), 0, 5).lastBatch()
        self.assertEqual(
            list(iter(last)), ['six', 'seven', 'eight', 'nine', 'ten'])
        # last batch when we are in the second set of items of the batch.
        last = _Batch(self.getData(), 5, 5).lastBatch()
        self.assertEqual(
            list(iter(last)), ['six', 'seven', 'eight', 'nine', 'ten'])
        # last batch when we get a request for an out of range item.
        last = _Batch(self.getData(), 99, 5).lastBatch()
        self.assertEqual(
            list(iter(last)), ['six', 'seven', 'eight', 'nine', 'ten'])

    def test_batchRoundTrip(self):
        batch = _Batch(self.getData(), 0, 3).nextBatch()
        self.assertEqual(list(iter(batch.nextBatch().prevBatch())),
                         list(iter(batch)))

    def test_first_last(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(batch.first(), 'one')
        self.assertEqual(batch.last(), 'three')
        batch = _Batch(self.getData(), 9, 3)
        self.assertEqual(batch.first(), 'ten')
        self.assertEqual(batch.last(), 'ten')
        batch = _Batch(self.getData(), 99, 3)
        self.assertRaises(IndexError, batch.first)
        self.assertRaises(IndexError, batch.last)

    def test_total(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(batch.total(), 10)
        batch = _Batch(self.getData(), 6, 3)
        self.assertEqual(batch.total(), 10)
        batch = _Batch(self.getData(), 99, 3)
        self.assertEqual(batch.total(), 10)

    def test_startNumber(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(batch.startNumber(), 1)
        batch = _Batch(self.getData(), 9, 3)
        self.assertEqual(batch.startNumber(), 10)
        batch = _Batch(self.getData(), 99, 3)
        self.assertEqual(batch.startNumber(), 100)

    def test_endNumber(self):
        batch = _Batch(self.getData(), 0, 3)
        self.assertEqual(batch.endNumber(), 3)
        batch = _Batch(self.getData(), 9, 3)
        self.assertEqual(batch.endNumber(), 10)
        batch = _Batch(self.getData(), 99, 3)
        self.assertEqual(batch.endNumber(), 100)


def additional_tests():
    return unittest.TestSuite((
        unittest.makeSuite(BatchTest),
        ))
